/*
all.hh -- header for geo
Copyright (C) 2003, 2004  Zsban Ambrus
*/

#include <utility> // pair
#include <list>
#include <vector>


// forward -----------------------------------------------------------------

class Value;
class token;


// mem.cc ------------------------------------------------------------------

class unleak {
	public:
	unleak (void);
	virtual ~unleak (void);
	};
	
#define UNLEAK private: unleak _unleak_instance;



// toke.ll -----------------------------------------------------------------

#define YY_DECL \
int yytoke (std::list<token*> *l)

YY_DECL;

extern FILE *yyin;


class token {
	protected: 
	token (void);
	virtual void I_want_runtime_class_information (void) { };
	};

class token_int:
	public token {
	public:
	token_int (long);
	long val;
	};

class token_punct:
	public token {
	public:
	token_punct (int);
	int chr;
	};

class token_sym:
	public token {
	UNLEAK
	public:
	token_sym (void);
	virtual ~token_sym (void);
	Value *val;
	};


void init_symtable (void);

void done_symtable (void);


// parse.yy ---------------------------------------------------------------

int yyparse (void);

void init_parse (std::list<token *> *tokens);


// value.cc ----------------------------------------------------------------


class Value {
	UNLEAK
	public:
	virtual long get_int (void);
	virtual bool get_bool (void);
	virtual Value *copy (void);
	virtual ~Value (void);
	protected: 
	Value (void);
	};

class Val_int:
	public Value {
	public:
	Val_int (long= 0);
	long get_int (void);
	bool get_bool (void);
	Value *copy (void);
	~Val_int (void);
	private:
	long val;
	};

class refcnt {
	UNLEAK
	public:
	void newref (void);
	void delref (void);
	virtual ~refcnt (void);
	protected:
	refcnt (void);
	private:
	int rcnt;
	};
	
class Data: 
	public refcnt {
	};

class Val_pv: /*SvPV*/
	public Value {
	public:
	Val_pv (Data *);
	bool get_bool (void);
	Value *copy (void);
	~Val_pv (void);
	Data *get (void);
	private:
	Data *const ptr;
	};

class Data_arr:
	public Data {
	public:
	Data_arr (void);
	void push (Value *);
	Value *aget (int);
	Value *aset (int, Value *);
	int length (void);
	~Data_arr (void);
	private:
	void extend (int);
	std::vector<Value *> vec;
	};


class Lvalue {
	UNLEAK
	public:
	virtual Value *value (void);
	virtual Value *set (Value *);
	virtual ~Lvalue (void);
	virtual Lvalue *copy (void);
	protected:
	Lvalue (void);
	};
	
class Lval_var:
	public Lvalue {
	public:
	Lval_var (token_sym *);
	Value *value (void);
	Value *set (Value *);
	~Lval_var (void);
	Lvalue *copy (void);
	private:
	token_sym *const ptr;
	};

class Lval_ind:
	public Lvalue {
	public:
	Lval_ind (Val_pv *, int);
	Value *value (void);
	Value *set (Value *);
	~Lval_ind (void);
	Lvalue *copy (void);
	private:
	Val_pv *const arr;
	int ind;
	};

class In_llist:
	public refcnt {
	public:
	In_llist (void);
	~In_llist (void);
	private:
	std::vector<Lvalue *> vec;
	friend class Lval_list;
	};

class Lval_list:
	public Lvalue {
	public:
	Lval_list (void);
	~Lval_list (void);
	void push (Lvalue *);
	Value *value (void);
	Value *set (Value *);
	Lvalue *copy (void);
	private:
	Lval_list (In_llist *);
	In_llist *const arr;
	};



// code.cc ----------------------------------------------------------------

int yylex (void);

void unlex (int c);

/* 
The block_? functions serve two purposes: dynamic scoping and loop control.
*/

void block_begin (void);

void block_end (void);

void block_redo (void);

void block_local (Lvalue *);

typedef std::list<token *>::iterator code_iterator_t;

class Val_lambda:
	public Value {
	public:
	Val_lambda (void);
	~Val_lambda (void);
	Val_lambda *copy (void);
	friend void block_enter (Val_lambda *l);
	private:
	Val_lambda (Val_lambda&);	
	code_iterator_t pos;
	friend void block_call (Val_lambda *l);
	};

void block_call (Val_lambda *l);



// ------------------------------------------------------------------------

;

//END
