/*
parse.yy -- grammar rules for geo
Copyright (C) Zsban Ambrus 2003, 2004
*/


%{

#include "all.hh"
#include <iostream>
#include <assert.h>

using std::cerr;


void yyerror (char *);
int yylex (void);

%}

%union {
Value *v;
Val_pv *arr;
Lvalue *l;
Lval_list *ll;
Lval_var *var;
}


/* Ha ujfajta term. szimbolumot talalsz ki, rakd hozza a skipsimple-hoz is! */

%token <v> T_INT
%token <var> T_SYM

%token '?' '!' T_CTRL_SKIP

%nonassoc T_PRINT
%right '=' T_PLUSLET T_MINUSLET T_TIMESLET T_DIVIDESLET T_MODULUSLET
%left T_EQ '<' T_NE T_LE '>' T_GE
%left '+' '-' 
%left '*' '/' '%'
%nonassoc T_PLUSPLUS T_MINUSMINUS
%nonassoc '['
%nonassoc T_VAR
%token T_LET

%type <v> expr rexpr stmt stmtl
%type <arr> arr_expr arr_expre
%type <l> lexpr lexpr1
%type <ll> arr_lexpr arr_lexprc arr_lexpre
//%type stmtc cond skip skiph skipsimple

%%



input	
	: { block_begin (); } stmtl { delete $2; block_end (); }
	| /* epsilon */
	;


/* ide jobb lenne egy balelemz */

cond
	: stmt { if ($1->get_bool ()) unlex (T_CTRL_SKIP); delete $1; } '?'
	| stmt { if (!$1->get_bool ()) unlex (T_CTRL_SKIP); delete $1; } '!'
	;

stmtl
	: stmt
	| cond T_CTRL_SKIP expr ';' skip { $$= $3; }
	| cond T_CTRL_SKIP ';' skip { $$= new Val_int; }
	;

stmt	
	: expr
	| stmtc { $$= new Val_int; }
	| stmtc expr { $$= $2; }
	;
	
stmtc
	: ';'
	| stmt ';' { delete $1; }
	| cond skiph ';'
	| stmtc '*' ')' { block_redo (); }
	;

expr
	: lexpr { $$= $1->value (); delete $1; }
	| rexpr
	;
	
rexpr
	: { block_begin (); } '(' stmtl ')' { $$= $3; block_end (); }
	| T_INT
	| '[' arr_expre ']' { $$= $2; }
	| T_PRINT expr { std::cout <<$2->get_int () <<'\n'; delete $2; $$= new Val_int (1); }
	| lexpr '=' expr { $$= $1->set ($3); delete $1; }
	| lexpr T_PLUSLET expr { 
		Value *p= $1->value ();
		$$= $1->set (new Val_int (
			p->get_int ()+$3->get_int ()
			)); 
		delete p; delete $1; delete $3; 
		}
	| lexpr T_MINUSLET expr { $$= $1->set (new Val_int ($1->value ()->get_int ()-$3->get_int ())); delete $1; delete $3; }
	| lexpr T_TIMESLET expr { $$= $1->set (new Val_int ($1->value ()->get_int ()*$3->get_int ())); delete $1; delete $3; }
	| lexpr T_DIVIDESLET expr { 
		$$= $1->set (new Val_int (
			$3->get_int () ? $1->value ()->get_int ()/$3->get_int () : 42
			)); 
		delete $1; delete $3; 
		}
	| lexpr T_MODULUSLET expr { $$= $1->set (new Val_int ($3->get_int () ? $1->value ()->get_int ()%$3->get_int () : 0)); delete $1; delete $3; }
	| expr T_EQ expr { $$= new Val_int ($1->get_int()==$3->get_int()); delete $1; delete $3; }
	| expr '<' expr { $$= new Val_int ($1->get_int()<$3->get_int()); delete $1; delete $3; }
	| expr T_NE expr { $$= new Val_int ($1->get_int()!=$3->get_int()); delete $1; delete $3; }
	| expr T_LE expr { $$= new Val_int ($1->get_int()<=$3->get_int()); delete $1; delete $3; }
	| expr '>' expr { $$= new Val_int ($1->get_int()>$3->get_int()); delete $1; delete $3; }
	| expr T_GE expr { $$= new Val_int ($1->get_int()>=$3->get_int()); delete $1; delete $3; }
	| expr '+' expr { 
		$$= new Val_int ($1->get_int()+$3->get_int()); 
		delete $1; delete $3; 
		}
	| expr '-' expr { $$= new Val_int ($1->get_int()-$3->get_int()); delete $1; delete $3; }
	| '+' expr { $$= $2; }
	| '-' expr { $$= new Val_int (-$2->get_int ()); delete $2; }
	| expr '*' expr { $$= new Val_int ($1->get_int()*$3->get_int()); delete $1; delete $3; }
	| expr '/' expr { 
		$$= new Val_int (
			$3->get_int() ? $1->get_int()/$3->get_int() : 42
			); /* LONG_MIN/(-1) ? */
		delete $1; delete $3; }
	| expr '%' expr { 
		$$= new Val_int (
			$3->get_int() ? $1->get_int()%$3->get_int() : 0
			);
		delete $1; delete $3; }
	| lexpr T_PLUSPLUS {
		Value *p= $1->value ();
		$$= $1->set (new Val_int (p->get_int ()+1));
		delete p; delete $1;
		}
	| lexpr T_MINUSMINUS { Value *p= $1->value (); $$= $1->set (new Val_int (p->get_int ()-1)); delete p; delete $1; }
	| T_PLUSPLUS lexpr1 {
		$$= $2->value ();
		delete $2->set (new Val_int ($$->get_int ()+1));
		delete $2;
		}
	| T_MINUSMINUS lexpr1 { $$= $2->value (); delete $2->set (new Val_int ($$->get_int ()-1)); delete $2; }
	;

/* Ha a tmb csupa balrtkbl ll, akkor nem az arr_expr->arr_lexprc rexpr
   szably alaktja jobbrtkk, hanem az expr->lexpr. */

arr_expre
	: arr_expr { $$= $1; }
	| arr_expr ',' { $$= $1; }
	;

arr_expr
	: arr_lexprc rexpr {
		$$= dynamic_cast<Val_pv *>($1->value ()); assert ($$); 
		delete $1; 
		dynamic_cast<Data_arr &>(*$$->get ()).push ($2);
		}
	| arr_expr ',' expr {
		$$= $1; dynamic_cast<Data_arr &>(*$$->get ()).push ($3);
		}
	;


// balrtkek 

/*
A lexpr1 szimblumot azrt vezetjk be, mert a	++(longexpr)[expr]
alak kifejezst csak gy lehetne LR1 mdon parsolni, ha az lvaluet s
az rvaluet egytt kezelnnk.
*/

lexpr1
	: T_SYM { $$= $1; }
	| '[' arr_lexpre ']' { $$= $2; }
	| T_VAR T_SYM { block_local ($2->copy ()); $$= $2; }
	| T_VAR '[' arr_lexpre ']' { block_local ($3->copy ()); $$= $3; }
	;
	
lexpr
	: lexpr1
	| expr '[' expr ']' {
		if (!dynamic_cast<Val_pv *>($1)) {
			cerr <<"error: indexing non-pv value\n";
			exit (1);
			}
		$$= new Lval_ind (dynamic_cast<Val_pv *>($1), $3->get_int ());
		delete $3;
		}
	;

arr_lexpre
	: arr_lexpr { $$= $1; }
	| arr_lexprc { $$= $1; }
	;

arr_lexpr
	: arr_lexprc lexpr { $$= $1; $$->push ($2); }
	;

arr_lexprc
	: /* epsilon */ { $$= new Lval_list; }
	| arr_lexpr ','
	;



/* most jn a legviccesebb rsz */

skiph
	: /* epsilon */
	| skiph skipsimple
	| skiph '(' skip ')'
	| skiph '[' skip ']'
	| skiph '{' skip '}'
	;
	
skip
	: /* epsilon */
	| skip skipsimple
	| skip ';'
	| skip '(' skip ')'
	| skip '[' skip ']'
	| skip '{' skip '}'
	;
	
skipsimple
	: T_INT { delete $1; } 
	| T_SYM { delete $1; }
	| T_PRINT | T_VAR | T_LET
	| '+' | '-' | '*' | '/' | '%' | T_EQ | '<' | T_NE | T_LE | '>' | T_GE
	| '=' | T_PLUSLET | T_MINUSLET | T_TIMESLET | T_DIVIDESLET | T_MODULUSLET
	| T_PLUSPLUS | T_MINUSMINUS
	| ',' | '?' | '!'
	;


%%


void yyerror (char *s) {
	cerr <<s <<'\n';
	}




//END
