/*
toke.ll -- lexer rules
Copyright (C) Zsban Ambrus 2003, 2004
*/

%{

#include "all.hh"
#include "parse.tab.hh"
#include <list>
#include <iostream>
#include <string>
#include <utility>
#include <map>

using std::cerr;


/*
#define YY_DECL \
int yytoke (std::list<token*> *l)
*/

typedef std::map <std::string, token*> symtable_t;

symtable_t symtable;

static token *toke_id (char *, int);

%}



%pointer

%x	COMM

FLOAT	([0-9]+(\.[0-9]*)?)([eE][-+]?[0-9]+)?

%%

[0-9]+	{
	long d;
	if (1>sscanf (yytext, "%ld", &d)) {
		cerr << "internal error: wrong decimal integer\n";
		return 0;
		}
	else {
		l->push_back (new token_int (d));
		}
	}

[+\-*/%=<>;,()\[\]{}?!] {
	l->push_back (new token_punct(*yytext));
	}

\+\+	{
	l->push_back (new token_punct(T_PLUSPLUS));
	}
--	{
	l->push_back (new token_punct(T_MINUSMINUS));
	}
	
==	{
	l->push_back (new token_punct(T_EQ));
	}
!=	{
	l->push_back (new token_punct(T_NE));
	}
\<=	{
	l->push_back (new token_punct(T_LE));
	}
\>=	{
	l->push_back (new token_punct(T_GE));
	}

\+=	{
	l->push_back (new token_punct(T_PLUSLET));
	}
-=	{
	l->push_back (new token_punct(T_MINUSLET));
	}
\*=	{
	l->push_back (new token_punct(T_TIMESLET));
	}
\/=	{
	l->push_back (new token_punct(T_DIVIDESLET));
	}
\%=	{
	l->push_back (new token_punct(T_MODULUSLET));
	}


[a-zA-Z_][a-zA-Z_0-9]*	{
	l->push_back (toke_id (yytext, yyleng));
	}
	

[[:space:]]	/*NOOP*/

#.*		/*NOOP*/
\/\/.*		//NOOP

"/*"		BEGIN(COMM);  /* Handling c-like comments: code after the example in flex */
<COMM>[^*\n]*	/*NOOP*/
<COMM>\*	/*NOOP*/
<COMM>\n	/*NOOP*/
<COMM>"*/"      BEGIN(INITIAL);
<COMM><<EOF>>	{
	cerr <<"error: eof in comment\n";
	exit (1);
	}


[[:print:]]		{
	cerr <<"error: wrong char " <<*yytext <<'\n';
	exit (1);
	}
.		{
	cerr <<"error: wrong char " <<std::oct <<*yytext <<'\n';
	exit (1);
	}

<<EOF>>		{
	yyterminate ();
	}

%%


static token *toke_id (char *n, int l) {
	std::pair<symtable_t::iterator, bool> p;
	p= symtable.insert (
		symtable_t::value_type (std::string (n, l), 0)
		);
	if (p.second)
		return p.first->second= new token_sym ();
	else
		return p.first->second;
	}


int yywrap (void) {
	return 1;
	}

void init_symtable (void) {
	symtable ["print"]= new token_punct (T_PRINT);
	symtable ["var"]= new token_punct (T_VAR);
	}

static void done_symbol (symtable_t::value_type &t) {
	if (dynamic_cast<token_sym *>(t.second))
		delete dynamic_cast<token_sym *>(t.second);
	}

void done_symtable (void) {
	for_each (symtable.begin (), symtable.end (), &done_symbol);
	symtable= symtable_t ();
	}

token::token (void) {
	}

token_int::token_int (long n):
	val (n) {
	}

token_punct::token_punct (int c):
	chr (c) {
	}
	
token_sym::token_sym (void):
	val (new Val_int) {
	}

token_sym::~token_sym (void) {
	delete val;
	}



//END
