/*
value.cc -- methods for values and symbols
Copyright (C) 2003, 2004  Zsban Ambrus
*/


#include "all.hh"
#include <iostream>

using std::cerr;




// -------------------------------------------------------------- value

Value::Value (void) {
	}

Value::~Value (void) {
	}

long Value::get_int (void) {
	cerr <<"error: integer type mismatch\n";
	exit (1);
	}

bool Value::get_bool (void) {
	return 1;
	}

Value *Value::copy (void) {
	cerr <<"internal error: cannot copy abstract value\n";
	exit (1);
	}


Val_int::Val_int (long l/*= 0*/) {
	val= l;
	}

Val_int::~Val_int (void) {
	}

Value *Val_int::copy (void) {
	return new Val_int (val);
	}

long Val_int::get_int (void) {
	return val;
	}

bool Val_int::get_bool (void) {
	return val!=0;
	}


/* beware: you have to newref immediately */
refcnt::refcnt (void):
	rcnt (0) {
	}

void refcnt::newref (void) {
	++rcnt;
	}
	
/* 
Vajon ez a megvalstsa a refcountingnak szablyos?  Elg biztos vagyok
benne, hogy mindig mkdik, de nem biztos, hogy szp dolog trlni
egy olyan objektumot, aminek a metdusa a trls utn mg fut, noha az
objektumra mr nem hivatkozik.  Taln szebb lenne csak 1-et visszaadni,
s a delref minden hvsnl ellenrizni, s ha 1 a visszatrsi rtk,
trlni a refcntt.
*/
void refcnt::delref (void) {
	if (!--rcnt)
		delete this;
	}

refcnt::~refcnt (void) {
	}


Val_pv::Val_pv (Data *what):
	ptr (what) {
	what->newref ();
	}
	
Value *Val_pv::copy (void) {
	return new Val_pv (ptr);
	}

Val_pv::~Val_pv (void) {
	ptr->delref ();
	}
	
bool Val_pv::get_bool (void) {
	if (dynamic_cast<Data_arr *>(ptr))
		return 0!=dynamic_cast<Data_arr *>(ptr)->length ();
	return 1;
	}

Data *Val_pv::get (void) {
	return ptr;
	}


Data_arr::Data_arr (void) {
	}

void Data_arr::extend (int g) {
	if (g<0) {
		cerr <<"error: negative array index\n";
		exit (1);
		}
	int s= vec.size ();
	if (g<s)
		return;
	vec.reserve (g);
	while (s++<=g) {
		vec.push_back (new Val_int);
		}
	}

Value *Data_arr::aget (int i) {
	extend (i);
	return vec[i]->copy ();
	}
	
Value *Data_arr::aset (int i, Value *u) {
	extend (i);
	delete vec[i];
	vec[i]= u->copy ();
	return u;
	}

int Data_arr::length (void) {
	return vec.size ();
	}

/* note that this function does not copy the value */
void Data_arr::push (Value *v) {
	vec.push_back (v);
	}

static inline void done_val (Value *v) {
	delete v;
	}

Data_arr::~Data_arr (void) {
	/* cerr <<"[Del arr]\n"; */
	for_each (vec.begin (), vec.end (), &done_val);
	}


// ------------------------------------------------------- lvalue

Lvalue::Lvalue (void) {
	}

Lvalue::~Lvalue (void) {
	}

/*
Copying an lvalue is needed when used with the "var" keyword: you need a copy
to assign to if "var" is used with an init'izer, and one to save in the dynamic
scope.  See parse.yy.
*/
Lvalue *Lvalue::copy (void) {
	cerr <<"internal error: copying abstract lvalue\n";
	exit (1);
	}

Value *Lvalue::value (void) {
	cerr <<"internal error: value of abstract lvalue\n";
	exit (1);
	}

Value *Lvalue::set (Value *d) {
	cerr <<"internal error: setting abstract lvalue\n";
	exit (1);
	}

Lval_var::Lval_var (token_sym *s):
	ptr (s) {
	}

Lval_var::~Lval_var (void) {
	}

Value *Lval_var::value (void) {
	return ptr->val->copy ();
	}

Value *Lval_var::set (Value *v) {
	delete ptr->val;
	ptr->val= v->copy ();
	return v;
	}

Lvalue *Lval_var::copy (void) {
	return new Lval_var (ptr);
	}


Lval_ind::Lval_ind (Val_pv *a, int j):
	arr (a), ind (j) {
	if (!dynamic_cast<Data_arr *>(arr->get ())) {
		cerr <<"error: subscripting non-array value\n";
		exit (1);
		}
	}
	
/*
Feladat: tanulmnyozza a parsert, s dntse el, okozhat-e gondot az,
hogy az Lvalue::value fggvnyt tl ksn hvjuk meg, gy ksbbi
retket kapunk, mint kne.  Ez nem csak az Lval_ind-re vonatkozik,
hanem brmelyik msik Lvalue osztly value fggvnyre.

Pldul tekintsk a kvetkez kdot:

  (var a= [4]; a [(a= [3]; 0)])

Ennek az eredmnye 4, noha az ember azt gondoln, hogy 3 lesz.
Szerintem ebben az esetben a program helyesen (noha non-intuitven)
viselkedik, hiszen szabadon eldntheti, hogy a [] opertor bal oldalt
rtkeli ki elszr.  Ezt a viselkedst adja a Perl is, ha ezt rjuk:

  do { $a= [4]; $$a[$a= 3, 0]; }

Dntse el, van-e olyan eset, amikor a ksi kirtkels hibs eredmnyre
vezet, s egyltaln, mi szmt hibs eredmnynek.
*/

Value *Lval_ind::value (void) {
	return dynamic_cast<Data_arr &>(*arr->get ()).aget (ind);
	}

Value *Lval_ind::set (Value *v) {
	return dynamic_cast<Data_arr &>(*arr->get ()).aset (ind, v);
	}

Lval_ind::~Lval_ind (void) {
	delete arr;
	}

Lvalue *Lval_ind::copy (void) {
	return new Lval_ind (dynamic_cast<Val_pv *>(arr->copy ()), ind);
	}


In_llist::In_llist (void) {
	}
	
static inline void done_lval (Lvalue *a) {
	delete a;
	}
	
In_llist::~In_llist (void) {
	for_each (vec.begin (), vec.end (), &done_lval);
	}

Lval_list::Lval_list (void):
	arr (new In_llist) {
	arr->newref ();
	}
	
Lval_list::~Lval_list (void) {
	arr->delref ();
	}
	
void Lval_list::push (Lvalue *l) {
	arr->vec.push_back (l);
	}
	
/* azt hiszem, ezt ssze lehetett volna rakni bind1st-bl, mem_fun1-bl, s
   hasonlkbl, de az olvashatatlanabb lenne. */
class func_push_value {
	public:
	func_push_value (Data_arr *p): arr (p) { }
	void operator() (Lvalue *l) { arr->push (l->value ()); }
	private:
	Data_arr *arr;
	};

Value *Lval_list::value (void) {
	Data_arr *r= new Data_arr;
	for_each (arr->vec.begin (), arr->vec.end (),
		func_push_value (r)
		);
	return new Val_pv (r);
	}
	
/* 
List-assignment semantics: a, b, c are lvalues; u is array [a, b, c]= u
EQUIV  (a= u[0]; b= u[1]; c= u[2]; u) Thus if right is longer, the full
array is returned, if right is shorter, the excess lefts are zeroed.  As a
special exception, []= u is legal (and is a noop) even if u is not an array.
*/
Value *Lval_list::set (Value *f) {
	int i, l;
	if (!arr->vec.size ())
		return f;
	if (!dynamic_cast<Val_pv *> (f)) {
		cerr <<"error: assigning single value to list\n";
		exit (1);
		}
	Data_arr *fa= 
		dynamic_cast<Data_arr *> (dynamic_cast<Val_pv *> (f)->get ());
	if (!fa) {
		cerr <<"error: assigning non-array value to list\n";
		exit (1);
		}
	l= std::min ((int)(arr->vec.size ()), fa->length ());
	Data_arr *r= new Data_arr;
	for (i= 0; i<l; i++)
		r->push (arr->vec [i]->set (fa->aget (i)));
	l= fa->length ();
	for (; i<l; i++)
		r->push (fa->aget (i));
	l= arr->vec.size ();
	for (; i<l; i++)
		delete arr->vec [i]->set (new Val_int);
	delete f;
	return new Val_pv (r);
	}

Lvalue *Lval_list::copy (void) {
	return new Lval_list (arr);
	}
	
Lval_list::Lval_list (In_llist *a):
	arr (a) {
	a->newref ();
	}





//END
